from chamber.patch import Options

import django.db.models.options as options
from django.db import models
from django.db.models.fields import (
    Field, URLField, AutoField, DateField, DateTimeField, DecimalField, GenericIPAddressField, IPAddressField,
    BooleanField, TextField, CharField, IntegerField, FloatField, SlugField, EmailField, NullBooleanField
)
from django.db.models.fields.related import ForeignKey, ManyToManyField, ForeignObjectRel

from pyston.utils.compatibility import get_last_parent_pk_field_name
from pyston.filters.default_filters import (
    BooleanFieldFilter, NullBooleanFieldFilter, StringFieldFilter, IntegerFieldFilter, FloatFieldFilter,
    DecimalFieldFilter, DateFilter, DateTimeFilter, GenericIPAddressFieldFilter, IPAddressFilterFilter,
    ManyToManyFieldFilter, ForeignKeyFilter, ForeignObjectRelFilter, CaseSensitiveStringFieldFilter
)


def merge_iterable(a, b):
    c = list(a) + list(b)
    return sorted(set(c), key=lambda x: c.index(x))


class RESTOptions(Options):

    model_class = models.Model
    meta_class_name = 'RESTMeta'
    meta_name = '_rest_meta'
    attributes = {}

    def __init__(self, model):
        super(RESTOptions, self).__init__(model)
        pk_field_name = get_last_parent_pk_field_name(model)

        fields = self._getattr('fields', ())

        self.default_fields = self._getattr('default_fields', (pk_field_name, '_obj_name'))
        self.detailed_fields = self._getattr('detailed_fields', fields)
        self.general_fields = self._getattr('general_fields', fields)
        self.direct_serialization_fields = self._getattr('direct_serialization_fields', fields)
        self.extra_fields = merge_iterable(
            self._getattr('extra_fields', ()),
            set(fields) - set(self.detailed_fields) - set(self.general_fields) - set(self.default_fields)
        )
        self.guest_fields = self._getattr('guest_fields', (pk_field_name, '_obj_name'))
        self.filter_fields = self._getattr('filter_fields', None)
        self.order_fields = self._getattr('order_fields', None)
        self.extra_filter_fields = self._getattr('extra_filter_fields', ())
        self.extra_order_fields = self._getattr('extra_order_fields', ())


options.DEFAULT_NAMES = options.DEFAULT_NAMES + ('default_fk_filter', 'default_m2m_filter', 'default_rel_filter')


def field_init(self, *args, **kwargs):
    _filter = kwargs.pop('filter', None)
    if _filter:
        self._filter = _filter
    self._init_pyston_tmp(*args, **kwargs)


def field_get_filter_class(self):
    return self._filter if hasattr(self, '_filter') else self.default_filter


def fk_get_filter_class(self):
    return (
        self._filter if hasattr(self, '_filter')
        else getattr(self.related_model._meta, 'default_fk_filter', None) or self.default_filter
    )


def m2m_get_filter_class(self):
    return (
        self._filter if hasattr(self, '_filter')
        else getattr(self.related_model._meta, 'default_m2m_filter', None) or self.default_filter
    )


def rel_get_filter_class(self):
    return getattr(self.field.model._meta, 'default_rel_filter', None) or self.default_filter


Field.default_filter = None
Field._init_pyston_tmp = Field.__init__
Field.__init__ = field_init
Field.filter = property(field_get_filter_class)

ForeignKey.filter = property(fk_get_filter_class)

ManyToManyField.filter = property(m2m_get_filter_class)

ForeignObjectRel.filter = property(rel_get_filter_class)

BooleanField.default_filter = BooleanFieldFilter
NullBooleanField.default_filter = NullBooleanFieldFilter
TextField.default_filter = StringFieldFilter
CharField.default_filter = StringFieldFilter
IntegerField.default_filter = IntegerFieldFilter
FloatField.default_filter = FloatFieldFilter
DecimalField.default_filter = DecimalFieldFilter
AutoField.default_filter = IntegerFieldFilter
DateField.default_filter = DateFilter
DateTimeField.default_filter = DateTimeFilter
GenericIPAddressField.default_filter = GenericIPAddressFieldFilter
IPAddressField.default_filter = IPAddressFilterFilter
ManyToManyField.default_filter = ManyToManyFieldFilter
ForeignKey.default_filter = ForeignKeyFilter
ForeignObjectRel.default_filter = ForeignObjectRelFilter
SlugField.default_filter = CaseSensitiveStringFieldFilter
EmailField.default_filter = CaseSensitiveStringFieldFilter



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import base64
import os

from django.conf import settings
from django.test.utils import override_settings
from django.utils.translation import ugettext

from germanium.anotations import data_provider
from germanium.tools.trivials import assert_in, assert_equal, assert_not_equal
from germanium.tools.http import assert_http_bad_request
from germanium.tools.rest import assert_valid_JSON_created_response, assert_valid_JSON_response

import responses

from pyston.conf import settings as pyston_settings

from .test_case import PystonTestCase

from app.models import User, Issue


class DataProcessorsTestCase(PystonTestCase):

    @data_provider('get_users_data')
    def test_create_user_with_file(self, number, data):
        data['contract'] = {
            'content_type': 'text/plain',
            'filename': 'contract.txt',
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        data = self.deserialize(resp)
        assert_not_equal(data['contract'], None)
        assert_in('filename', data['contract'])
        assert_in('url', data['contract'])
        assert_equal(data['contract']['content_type'], 'text/plain')

    @data_provider('get_users_data')
    def test_create_user_with_file_and_not_defined_content_type(self, number, data):
        data['contract'] = {
            'filename': 'contract.txt',
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        data = self.deserialize(resp)
        assert_not_equal(data['contract'], None)
        assert_in('filename', data['contract'])
        assert_in('url', data['contract'])
        assert_equal(data['contract']['content_type'], 'text/plain')

    @data_provider('get_users_data')
    def test_error_during_creating_user_with_file_and_not_defined_content_type(self, number, data):
        data['contract'] = {
            'filename': 'contract',
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)

    def serve_file(self, rsps, url, body, status=200):
        rsps.add(responses.GET, url, body=body, status=status, match_querystring=True)

    def get_file_url_response(self, data):
        url = 'http://foo.bar/testfile.pdf'
        data['contract'] = {
            'filename': 'testfile.pdf',
            'url': url,
        }
        with responses.RequestsMock(assert_all_requests_are_fired=True) as rsps:
            with open(os.path.join(settings.PROJECT_DIR, 'data', 'tests', 'pdf-sample.pdf'), 'rb') as f:
                content = f.read()

            self.serve_file(rsps, url, content)
            return self.post(self.USER_API_URL, data=data)

    @data_provider('get_users_data')
    def test_create_user_with_file_url(self, number, data):
        resp = self.get_file_url_response(data)

        assert_valid_JSON_created_response(resp)
        data = self.deserialize(resp)

        assert_not_equal(data['contract'], None)
        assert_in('filename', data['contract'])
        assert_in('url', data['contract'])
        assert_equal(data['contract']['content_type'], 'application/pdf')

    @override_settings(PYSTON_FILE_SIZE_LIMIT=7000)
    @data_provider('get_users_data')
    def test_create_user_with_file_url_too_large(self, number, data):
        resp = self.get_file_url_response(data)

        assert_http_bad_request(resp)
        assert_in('contract', self.deserialize(resp).get('messages', {}).get('errors', {}))

    @data_provider('get_issues_and_users_data')
    def test_atomic_create_issue_with_user_id(self, number, issue_data, user_data):
        resp = self.post(self.USER_API_URL, data=user_data)
        issue_data['created_by'] = self.get_pk(resp)
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_valid_JSON_created_response(resp)

    @data_provider('get_issues_data')
    def test_atomic_create_issue_with_user(self, number, data):
        users_before_count = User.objects.all().count()
        resp = self.post(self.ISSUE_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 2, User.objects.all().count())

    @data_provider('get_issues_data')
    def test_atomic_update_issue_with_user(self, number, data):
        users_before_count = User.objects.all().count()
        resp = self.post(self.ISSUE_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 2, User.objects.all().count())
        data['created_by'] = self.get_user_data()
        data['created_by']['id'] = self.deserialize(resp)['created_by']['id']
        data['leader'] = self.get_user_data()
        resp = self.post(self.ISSUE_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 3, User.objects.all().count())

    @data_provider('get_issues_and_users_data')
    def test_atomic_set_issue_with_user_reverse(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        user_data['createdIssues'] = {'set': (issue_data,)}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count + 1, Issue.objects.all().count())

    @data_provider('get_issues_and_users_data')
    def test_atomic_add_and_delete_issues_with_reverse(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()

        user_data['createdIssues'] = {'add': (self.get_issue_data(), self.get_issue_data(), self.get_issue_data())}
        resp = self.post(self.USER_API_URL, data=user_data)

        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count + 3, Issue.objects.all().count())
        user_pk = self.get_pk(resp)

        first_issue_data = self.get_issue_data()
        first_issue_data['solver'] = {'email': 'solver@email.cz', 'createdIssues': [self.get_issue_data()]}

        user_data['createdIssues'] = {'set': (first_issue_data, self.get_issue_data(), self.get_issue_data())}
        resp = self.put('%s%s/' % (self.USER_API_URL, user_pk), data=user_data)
        assert_equal(issues_before_count + 4, Issue.objects.all().count())
        assert_valid_JSON_response(resp)
        user_data['createdIssues'] = {'remove': list(Issue.objects.filter(created_by=user_pk).
                                                     values_list('pk', flat=True))}
        resp = self.put('%s%s/' % (self.USER_API_URL, self.get_pk(resp)), data=user_data)
        assert_valid_JSON_response(resp)
        assert_equal(issues_before_count + 1, Issue.objects.all().count())

        user_data['createdIssues'] = (self.get_issue_data(), self.get_issue_data(), self.get_issue_data())
        resp = self.put('%s%s/' % (self.USER_API_URL, user_pk), data=user_data)
        assert_equal(issues_before_count + 4, Issue.objects.all().count())
        assert_valid_JSON_response(resp)

    @data_provider('get_issues_and_users_data')
    def test_atomic_add_issues_by_m2m_reverse(self, number, issue_data, user_data):
        user_data['watchedIssues'] = {'add': (self.get_issue_data(), self.get_issue_data(), self.get_issue_data())}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_valid_JSON_created_response(resp)
        watched_issues = self.deserialize(resp)['watchedIssues']
        assert_equal(len(watched_issues), 3)
        assert_equal(Issue.objects.all().count(), 3)
        watched_issues_ids = [watched_issue['id'] for watched_issue in watched_issues]

        for issue in Issue.objects.all():
            assert_equal(list(issue.watched_by.values_list('email', flat=True)), [user_data['email']])

        user_data2 = self.get_user_data()
        user_data2['watchedIssues'] = watched_issues_ids
        resp = self.post(self.USER_API_URL, data=user_data2)
        assert_valid_JSON_created_response(resp)
        watched_issues = self.deserialize(resp)['watchedIssues']
        assert_equal(len(watched_issues), 3)
        assert_equal(Issue.objects.all().count(), 3)
        for issue in Issue.objects.all():
            assert_equal(list(issue.watched_by.values_list('email', flat=True)),
                              [user_data['email'], user_data2['email']])

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_REVERSE_FIELDS=False)
    def test_atomic_add_and_delete_issues_with_auto_reverse_fields_turned_off(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()

        user_data['createdIssues'] = {'add': (self.get_issue_data(), self.get_issue_data(), self.get_issue_data())}
        resp = self.post(self.USER_API_URL, data=user_data)

        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count, Issue.objects.all().count())

    @data_provider('get_issues_data')
    @override_settings(PYSTON_AUTO_RELATED_DIRECT_FIELDS=False)
    def test_atomic_create_issue_with_user_with_auto_related_fields_turned_off(self, number, data):
        users_before_count = User.objects.all().count()
        resp = self.post(self.ISSUE_API_URL, data=data)
        assert_http_bad_request(resp)

    @data_provider('get_issues_and_users_data')
    def test_atomic_add_delete_and_set_issues_with_errors(self, number, issue_data, user_data):
        user_data['createdIssues'] = {'set': (None, '', None, {}, [None])}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('set', self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues', {}))

        user_data['createdIssues'] = {'add': (None, '', None, [], {}, {'id': 500}),
                                      'remove': (None, '', None, {}, {'id': 500}, [])}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('add', self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues', {}))
        assert_in('remove', self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues', {}))

        user_data['createdIssues'] = {'add': None, 'remove': None}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('add', self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues', {}))
        assert_in('remove', self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues', {}))

        user_data['createdIssues'] = (self.get_issue_data(), self.get_issue_data(), 'invalid')
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_equal(
            self.deserialize(resp).get('messages', {}).get('errors', {}).get('createdIssues')[0]['_index'], 2
        )
        assert_http_bad_request(resp)

    @data_provider('get_issues_and_users_data')
    def test_atomic_set_issue_with_watchers(self, number, issue_data, user_data):
        users_before_count = User.objects.all().count()

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = self.get_users_data(flat=True)
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 12, User.objects.all().count())

        issue_data['leader'] = self.get_user_data()
        issue_data['created_by'] = self.get_user_data()
        issue_data['watched_by'] = {'set': self.get_users_data(flat=True)}

        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 24, User.objects.all().count())

    @data_provider('get_issues_and_users_data')
    def test_atomic_add_and_delete_issue_with_watchers(self, number, issue_data, user_data):
        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'add': self.get_users_data(flat=True)}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(len(self.deserialize(resp).get('watched_by')), 10)

        pk = self.get_pk(resp)
        issue_data['created_by'] = self.get_user_data()
        issue_data['leader'] = self.get_user_data()
        issue_data['watched_by'] = {'add': self.get_users_data(flat=True),
                                    'remove': [obj.get('id') for obj in self.deserialize(resp).get('watched_by')][:5]}
        resp = self.put('%s%s/' % (self.ISSUE_API_URL, pk), data=issue_data)
        assert_equal(len(self.deserialize(resp).get('watched_by')), 15)

        issue_data['watched_by'] = self.get_users_data(flat=True)
        issue_data['created_by'] = self.get_user_data()
        issue_data['leader'] = self.get_user_data()
        resp = self.put('%s%s/' % (self.ISSUE_API_URL, pk), data=issue_data)
        assert_equal(len(self.deserialize(resp).get('watched_by')), 10)

    @data_provider('get_issues_and_users_data')
    def test_atomic_add_delete_and_set_issue_with_watchers_with_errors(self, number, issue_data, user_data):
        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'add': [None, [], 'invalid_text', {}], 'remove': ['invalid_text', 5, {}]}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_in('add', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))
        assert_in('remove', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'set': [None, [], 'invalid_text', {}]}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_in('set', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'set': ''}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_in('set', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'set': None}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_in('set', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = {'set': {}}
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_in('set', self.deserialize(resp).get('messages', {}).get('errors').get('watched_by'))

        issue_data['created_by'] = user_data
        issue_data['watched_by'] = [{'invalid': 'invalid2'}]
        resp = self.post(self.ISSUE_API_URL, data=issue_data)
        assert_http_bad_request(resp)
        assert_equal(self.deserialize(resp).get('messages', {}).get('errors')['watched_by'][0]['_index'], 0)

    @data_provider('get_issues_and_users_data')
    def test_create_issue_via_user_one_to_one(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        issue_data['created_by'] = self.get_user_data()
        user_data['leading_issue'] = issue_data
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count + 1, Issue.objects.all().count())

        pk = self.get_pk(resp)
        user_data = {'leading_issue': None}
        resp = self.patch('%s%s/' % (self.USER_API_URL, pk), data=user_data)
        assert_valid_JSON_response(resp)
        assert_equal(issues_before_count, Issue.objects.all().count())

        user_data = {'leading_issue': None}
        resp = self.patch('%s%s/' % (self.USER_API_URL, pk), data=user_data)
        assert_valid_JSON_response(resp)
        assert_equal(issues_before_count, Issue.objects.all().count())

        user_data = {'leading_issue': self.get_issue_data()}
        resp = self.patch('%s%s/' % (self.USER_API_URL, pk), data=user_data)
        assert_valid_JSON_response(resp)
        assert_equal(issues_before_count + 1, Issue.objects.all().count())

    @data_provider('get_issues_and_users_data')
    def test_create_issue_via_user_one_to_one_bad_request(self, number, issue_data, user_data):
        issue_data['created_by'] = self.get_user_data()
        user_data['leading_issue'] = {}
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('leading_issue', self.deserialize(resp).get('messages', {}).get('errors'))

        user_data['leading_issue'] = 'bad data'
        resp = self.post(self.USER_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('leading_issue', self.deserialize(resp).get('messages', {}).get('errors'))

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_REVERSE_FIELDS=False)
    def test_reverse_with_defined_field_created_issues_renamed(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        user_data['created_issues_renamed'] = (self.get_issue_data(), self.get_issue_data(), self.get_issue_data())
        resp = self.post(self.USER_WITH_FORM_API_URL, data=user_data)

        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count + 3, Issue.objects.all().count())

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_DIRECT_FIELDS=False)
    def test_create_user_via_renamed_field(self, number, issue_data, user_data):
        users_before_count = User.objects.all().count()
        issue_data['created_by'] = self.get_user_data()
        issue_data['leader'] = self.get_user_data()
        resp = self.post(self.ISSUE_WITH_FORM_API_URL, data=issue_data)

        assert_valid_JSON_created_response(resp)
        assert_equal(users_before_count + 2, User.objects.all().count())

    @data_provider('get_issues_and_users_data')
    def test_reverse_with_defined_field_created_issues_renamed_fail(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        user_data['created_issues_renamed'] = {'add': (self.get_issue_data(),
                                                       self.get_issue_data(), self.get_issue_data())}
        resp = self.post(self.USER_WITH_FORM_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('created_issues_renamed', self.deserialize(resp).get('messages', {}).get('errors'))
        assert_equal(issues_before_count + 0, Issue.objects.all().count())

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_REVERSE_FIELDS=False)
    def test_reverse_with_defined_field_created_issues_renamed_invalid_issue_name(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        users_before_count = Issue.objects.all().count()
        user_data['created_issues_renamed'] = (
            self.get_issue_data(), self.get_issue_data(name='invalid'), self.get_issue_data()
        )
        resp = self.post(self.USER_WITH_FORM_API_URL, data=user_data)

        assert_http_bad_request(resp)
        assert_equal(issues_before_count, Issue.objects.all().count())
        assert_equal(users_before_count, User.objects.all().count())

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_REVERSE_FIELDS=False)
    def test_create_issue_via_user_one_to_one_renamed(self, number, issue_data, user_data):
        issues_before_count = Issue.objects.all().count()
        user_data['leading_issue_renamed'] = self.get_issue_data()
        resp = self.post(self.USER_WITH_FORM_API_URL, data=user_data)
        assert_valid_JSON_created_response(resp)
        assert_equal(issues_before_count + 1, Issue.objects.all().count())

    @data_provider('get_issues_and_users_data')
    @override_settings(PYSTON_AUTO_RELATED_REVERSE_FIELDS=False)
    def test_create_issue_via_user_one_to_one_renamed_fail(self, number, issue_data, user_data):
        user_data['leading_issue_renamed'] = {}
        resp = self.post(self.USER_WITH_FORM_API_URL, data=user_data)
        assert_http_bad_request(resp)
        assert_in('leading_issue_renamed', self.deserialize(resp).get('messages', {}).get('errors'))

    @data_provider('get_users_data')
    def test_should_raise_bad_request_with_invalid_filename(self, number, data):
        data['contract'] = {
            'filename': 'contract',
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        assert_equal(
            errors['contract'],
            ugettext('Content type cannot be evaluated from input data please send it')
        )

    @data_provider('get_users_data')
    def test_should_raise_bad_request_if_file_content_is_not_in_base64(self, number, data):
        data['contract'] = {
            'content_type': 'text/plain',
            'filename': 'contract.txt',
            'content': 'abc',
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        assert_equal(errors['contract']['content'], ugettext('File content must be in base64 format'))

    @data_provider('get_users_data')
    def test_should_raise_bad_request_if_url_is_not_valid(self, number, data):
        data['contract'] = {
            'filename': 'testfile.pdf',
            'url': 'hs://foo.bar/testfile.pdf',
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        assert_equal(errors['contract']['url'], ugettext('Enter a valid URL.'))

    @data_provider('get_users_data')
    def test_should_raise_bad_request_if_required_items_are_not_provided(self, number, data):
        REQUIRED_ITEMS = {'content'}
        REQUIRED_URL_ITEMS = {'url'}
        data['contract'] = {}
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        msg = ugettext('File data item must contains {} or {}').format(
                  ', '.join(REQUIRED_ITEMS), ', '.join(REQUIRED_URL_ITEMS)
              )
        assert_equal(errors['contract'], msg)

    @data_provider('get_users_data')
    def test_should_raise_bad_request_if_file_is_unreachable(self, number, data):
        url = 'http://foo.bar/testfile.pdf'
        data['contract'] = {
            'filename': 'testfile.pdf',
            'url': url,
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        assert_equal(errors['contract']['url'], ugettext('File is unreachable on the URL address'))

    @override_settings(PYSTON_FILE_SIZE_LIMIT=10)
    @data_provider('get_users_data')
    def test_should_raise_bad_request_if_response_is_too_large(self, number, data):
        data['contract'] = {
            'content_type': 'text/plain',
            'filename': 'contract.txt',
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.get_file_url_response(data)
        assert_http_bad_request(resp)
        errors = self.deserialize(resp).get('messages', {}).get('errors')
        assert_in('contract', errors)
        msg = ugettext('Response too large, maximum size is {} bytes').format(pyston_settings.FILE_SIZE_LIMIT)
        assert_equal(errors['contract']['url'], msg)

    @data_provider('get_users_data')
    def test_filename_and_content_type_can_be_evaluated_from_file_content(self, number, data):
        data['contract'] = {
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8')
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        assert_in('.txt', self.deserialize(resp)['contract']['filename'])

    @data_provider('get_users_data')
    def test_filename_override_evaluated_filename_from_content(self, number, data):
        data['contract'] = {
            'content': base64.b64encode(
                ('Contract of %s code: šří+áýšé' % data['email']).encode('utf-8')
            ).decode('utf-8'),
            'filename': 'test.csv'
        }
        resp = self.post(self.USER_API_URL, data=data)
        assert_valid_JSON_created_response(resp)
        assert_in('.csv', self.deserialize(resp)['contract']['filename'])



from django.db.models import Q
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext

from pyston.exception import RESTException
from pyston.utils import rfs, LOOKUP_SEP
from pyston.utils.helpers import get_field_or_none, get_method_or_none
from pyston.serializer import get_resource_or_none

from .exceptions import FilterValueError, OperatorFilterError, FilterIdentifierError
from .utils import LOGICAL_OPERATORS
from .parser import QueryStringFilterParser, DefaultFilterParser, FilterParserError


def get_allowed_filter_fields_rfs_from_model(model):
    return rfs(model._rest_meta.extra_filter_fields).join(rfs(model._rest_meta.filter_fields))


class ModelFilterManager:
    """
    Filter manager is used inside model resource for composing filters with purpose to restrict output data according
    to input values.
    This is abstract class that provides methods to obtain concrete filters from resource and model methods and fields.
    """

    def _get_method_filter(self, method, identifiers_prefix, identifiers, identifiers_suffix, model, resource, request,
                           filters_fields_rfs):
        """
        :param method: method from which we can get filter.
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param identifiers_suffix: list of suffixes that can be used for more specific filters.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter object that is obtained from method.
        """
        if hasattr(method, 'filter_by'):
            # If method has filter_by attribute filter is being searched according to this value.
            filter_by_identifiers = method.filter_by.split(LOOKUP_SEP)
            next_identifiers = filter_by_identifiers + identifiers_suffix
            # Because method must be inside allowed filter fields RFS, we must add value filter_by of the method
            # to the next RFS.
            next_filters_fields_rfs = rfs(filter_by_identifiers)
            return self._get_filter_recursive(
                identifiers_prefix, next_identifiers, model, resource, request, next_filters_fields_rfs
            )
        suffix = LOOKUP_SEP.join(identifiers_suffix)
        if not hasattr(method, 'filter') or (suffix and suffix not in method.filter.get_suffixes()):
            raise FilterIdentifierError
        return method.filter(identifiers_prefix, identifiers, identifiers_suffix, model, method=method)

    def _get_resource_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter that is obtained from resource and its methods.
        """
        # Filter is obtained from resource filters dict
        for i in range(1, len(identifiers) + 1):
            # Because resource filters can contains filter key with __ we must try all combinations with suffixes
            current_identifiers = identifiers[:i]
            identifiers_string = LOOKUP_SEP.join(current_identifiers)
            identifiers_suffix = identifiers[i:]
            suffix_string = LOOKUP_SEP.join(identifiers_suffix)
            if (resource and identifiers_string in resource.filters and
                    (not suffix_string or suffix_string in resource.filters[identifiers_string].get_suffixes())):
                return resource.filters[identifiers_string](
                    identifiers_prefix, current_identifiers, identifiers_suffix, model
                )

        # Filter is obtained from resource methods
        current_identifier = identifiers[0]
        resource_method = get_method_or_none(resource, current_identifier)

        if current_identifier in filters_fields_rfs and resource_method:
            return self._get_method_filter(
                resource_method, identifiers_suffix, [current_identifier], identifiers[1:], model, resource, request,
                filters_fields_rfs
            )

        return None

    def _get_model_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter from model fields or methods
        """
        current_identifier = identifiers[0]
        identifiers_suffix = identifiers[1:]

        if current_identifier not in filters_fields_rfs:
            raise FilterIdentifierError

        suffix = LOOKUP_SEP.join(identifiers_suffix)

        model_field = get_field_or_none(model, current_identifier)
        model_method = get_method_or_none(model, current_identifier)

        if model_field and model_field.filter and (not suffix or suffix in model_field.filter.get_suffixes()):
            return model_field.filter(
                identifiers_prefix, [current_identifier], identifiers_suffix, model, field=model_field
            )
        elif model_field and model_field.is_relation and model_field.related_model:
            # recursive search for filter via related model fields
            next_model = model_field.related_model
            next_resource = get_resource_or_none(request, next_model, getattr(resource, 'resource_typemapper', None))
            return self._get_filter_recursive(
                identifiers_prefix + [identifiers[0]], identifiers[1:], next_model, next_resource, request,
                filters_fields_rfs[current_identifier].subfieldset
            )
        elif model_method:
            return self._get_method_filter(
                model_method, identifiers_prefix, [current_identifier], identifiers_suffix,
                model, resource, request, filters_fields_rfs
            )

    def _get_filter_recursive(self, identifiers_prefix, identifiers, model, resource, request,
                              extra_filter_fields_rfs=None):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param extra_filter_fields_rfs: RFS of fields that is allowed to filter.
        :return: method search recursive filter with resource_filter and model_filter getters.
        """
        extra_filter_fields_rfs = rfs() if extra_filter_fields_rfs is None else extra_filter_fields_rfs
        filters_fields_rfs = (
            extra_filter_fields_rfs.join(
                resource.get_filter_fields_rfs() if resource else get_allowed_filter_fields_rfs_from_model(model)
            )
        )
        filter_obj = (
            self._get_resource_filter(
                identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs) or
            self._get_model_filter(
                identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs)
        )
        if not filter_obj:
            raise FilterIdentifierError
        return filter_obj

    def get_filter(self, identifiers, resource, request):
        """
        :param identifiers: list of identifiers that conclusively identifies a filter.
        :param resource: resource object.
        :param request: django HTTP request.
        :return: method returns filter object according to input identifiers, resource and request.
        """
        return self._get_filter_recursive([], identifiers, resource.model, resource, request)

    def filter(self, resource, qs, request):
        """
        :param resource: resource object.
        :param qs: model queryset that will be filtered.
        :param request: django HTTP request.
        :return: methods should return filtered queryset.
        """
        raise NotImplementedError


class ParserModelFilterManager(ModelFilterManager):
    """
    Manager that uses parser to parse input filter data to the logical conditions tree.
    """

    parser = None

    def _convert_logical_conditions(self, condition, resource, request):
        """
        Method that recursive converts condition tree to the django models Q objects.
        """
        if condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.NOT:
            return ~Q(self._convert_logical_conditions(condition.condition_right, resource, request))
        elif condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.AND:
            return Q(
                self._convert_logical_conditions(condition.condition_left, resource, request),
                self._convert_logical_conditions(condition.condition_right, resource, request)
            )
        elif condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.OR:
            return Q(
                self._convert_logical_conditions(condition.condition_left, resource, request) |
                self._convert_logical_conditions(condition.condition_right, resource, request)
            )
        else:
            try:
                return self.get_filter(condition.identifiers, resource, request).get_q(
                    condition.value, condition.operator_slug, request
                )
            except FilterIdentifierError:
                raise RESTException(
                    mark_safe(ugettext('Invalid identifier of condition "{}"').format(condition.source))
                )
            except FilterValueError as ex:
                raise RESTException(
                    mark_safe(ugettext('Invalid value of condition "{}". {}').format(condition.source, ex))
                )
            except OperatorFilterError:
                raise RESTException(
                    mark_safe(ugettext('Invalid operator of condition "{}"').format(condition.source))
                )

    def filter(self, resource, qs, request):
        try:
            parsed_conditions = self.parser.parse(request)
            return (
                qs.filter(pk__in=qs.filter(
                    self._convert_logical_conditions(parsed_conditions, resource, request)).values('pk')
                )
                if parsed_conditions else qs
            )
        except FilterParserError as ex:
            raise RESTException(ex)


class DefaultFilterManager(ParserModelFilterManager):
    """
    Default filter manager that provides complex logical operations with AND, OR, NOT operators.
    """

    parser = DefaultFilterParser()


class QueryStringFilterManager(ParserModelFilterManager):
    """
    Simple filter manager where one filter term is one query string. Terms are always joined with AND operator.
    """

    parser = QueryStringFilterParser()


class MultipleFilterManager(ModelFilterManager):
    """
    Helper that joins DefaultFilterManager and QueryStringFilterManager. Resource can use both options together.
    """

    def filter(self, resource, qs, request):
        return QueryStringFilterManager().filter(
            resource, DefaultFilterManager().filter(resource, qs, request), request
        )



from django.utils.translation import ugettext
from django.db.models.query import QuerySet

from .exception import RESTException


class BasePaginator:

    @property
    def page_qs(self):
        raise NotImplementedError

    @property
    def headers(self):
        return {}


class Paginator(BasePaginator):
    """
    REST paginator for list and querysets
    """

    MAX_BIG_INT = pow(2, 63) - 1

    def __init__(self, qs, request):
        self.qs = qs
        self.offset = self._get_offset(request)
        self.base = self._get_base(request)
        self.total = self._get_total()

    def _get_total(self):
        if isinstance(self.qs, QuerySet):
            return self.qs.count()
        else:
            return len(self.qs)

    def _get_offset(self, request):
        offset = request._rest_context.get('offset', '0')
        if offset.isdigit():
            offset_int = int(offset)
            if offset_int > self.MAX_BIG_INT:
                raise RESTException(ugettext('Offset must be lower or equal to {}').format(self.MAX_BIG_INT))
            else:
                return offset_int
        else:
            raise RESTException(ugettext('Offset must be natural number'))

    def _get_base(self, request):
        base = request._rest_context.get('base')
        if not base:
            return None
        elif base.isdigit():
            base_int = int(base)
            if base_int > self.MAX_BIG_INT:
                raise RESTException(ugettext('Base must lower or equal to {}').format(self.MAX_BIG_INT))
            else:
                return base_int
        else:
            raise RESTException(ugettext('Base must be natural number or empty'))

    @property
    def page_qs(self):
        if self.base is not None:
            return self.qs[self.offset:(self.offset + self.base)]
        else:
            return self.qs[self.offset:]

    @property
    def headers(self):
        return {'X-Total': self.total}



from django.conf import settings as django_settings


CONVERTERS = (
    'pyston.converters.JSONConverter',
    'pyston.converters.XMLConverter',
    'pyston.converters.CSVConverter',
)

DEFAULT_FILENAMES = (
    (('pdf', 'doc', 'docx', 'xls', 'xlsx', 'ppt', 'pptx', 'txt', 'csv'), 'document'),
    (('jpg', 'jpeg', 'png', 'gif', 'tiff', 'bmp', 'svg'), 'image'),
)

DEFAULT_FILENAME = 'attachment'

try:
    import xlsxwriter

    CONVERTERS += (
        'pyston.converters.XLSXConverter',
    )
except ImportError:
    pass

try:
    # pisa isn't standard with python. It shouldn't be required if it isn't used.
    from xhtml2pdf import pisa

    CONVERTERS += (
        'pyston.converters.PDFConverter',
    )
except ImportError:
    pass


DEFAULTS = {
    'CONVERTERS': CONVERTERS,
    'IGNORE_DUPE_MODELS': False,
    'CORS': False,
    'CORS_WHITELIST': (),
    'CORS_MAX_AGE': 60 * 30,
    'CORS_ALLOW_CREDENTIALS': True,
    'JSON_CONVERTER_OPTIONS': {
        'indent': 4
    },
    'PDF_EXPORT_TEMPLATE': 'default_pdf_table.html',
    'FILE_SIZE_LIMIT': 5000000,
    'AUTO_RELATED_REVERSE_FIELDS': True,
    'AUTO_RELATED_DIRECT_FIELDS': True,
    'PARTIAL_PUT_UPDATE': False,
    'PARTIAL_RELATED_UPDATE': False,
    'ERRORS_RESPONSE_CLASS': 'pyston.response.RESTErrorsResponse',
    'ERROR_RESPONSE_CLASS': 'pyston.response.RESTErrorResponse',
    'AUTO_REGISTER_RESOURCE': True,
    'ALLOW_TAGS': False,
    'DEFAULT_FILENAMES': DEFAULT_FILENAMES,
    'DEFAULT_FILENAME': DEFAULT_FILENAME,
}


class Settings:

    def __getattr__(self, attr):
        if attr not in DEFAULTS:
            raise AttributeError('Invalid Pyston setting: "{}"'.format(attr))

        return getattr(django_settings, 'PYSTON_{}'.format(attr), DEFAULTS[attr])


settings = Settings()



from django.utils.translation import ugettext
from django.utils.encoding import force_text

from .forms import RESTDictError, RESTDictIndexError, RESTListError, RESTValidationError


class HeadersResponse:

    fieldset = True

    def __init__(self, result, http_headers=None, code=200):
        http_headers = {} if http_headers is None else http_headers
        self.result = result
        self.http_headers = http_headers
        self.status_code = code


class NoFieldsetResponse(HeadersResponse):

    fieldset = False


class RESTResponse(NoFieldsetResponse):

    def __init__(self, msg, http_headers=None, code=200):
        http_headers = {} if http_headers is None else http_headers
        super(RESTResponse, self).__init__(result={'messages': msg}, http_headers=http_headers, code=code)


class RESTOkResponse(NoFieldsetResponse):

    def __init__(self, msg, http_headers=None, code=200):
        http_headers = {} if http_headers is None else http_headers
        super(RESTOkResponse, self).__init__(
            result={'messages': {'success': msg}}, http_headers=http_headers, code=code
        )


class RESTCreatedResponse(HeadersResponse):

    def __init__(self, result, http_headers=None, code=201):
        http_headers = {} if http_headers is None else http_headers
        super(RESTCreatedResponse, self).__init__(result=result, http_headers=http_headers, code=code)


class RESTNoContentResponse(NoFieldsetResponse):

    def __init__(self, http_headers=None, code=204):
        http_headers = {} if http_headers is None else http_headers
        super(RESTNoContentResponse, self).__init__(result='', http_headers=http_headers, code=code)


class RESTErrorsResponseMixin:

    def _get_errors(self, data):
        if isinstance(data, RESTDictIndexError):
            result = {
                '_index': data.index
            }
            result.update(self._get_errors(data.data))
            return result
        elif isinstance(data, (RESTDictError, dict)):
            return {
                key: self._get_errors(val) for key, val in data.items()
            }
        elif isinstance(data, (RESTListError, list, tuple)):
            return [self._get_errors(error) for error in data]
        else:
            return RESTValidationError(data).message


class RESTErrorsResponse(RESTErrorsResponseMixin, HeadersResponse):

    def __init__(self, msg, http_headers=None, code=400):
        http_headers = {} if http_headers is None else http_headers
        super(RESTErrorsResponse, self).__init__(
            result={'messages': {'errors': self._get_errors(msg)}}, http_headers=http_headers, code=code
        )


class RESTErrorResponse(RESTErrorsResponseMixin, NoFieldsetResponse):

    def __init__(self, msg, http_headers=None, code=400):
        http_headers = {} if http_headers is None else http_headers
        super(RESTErrorResponse, self).__init__(
            result={'messages': {'error': self._get_errors(msg)}}, http_headers=http_headers, code=code
        )


class ResponseFactory:

    def __init__(self, response_class):
        self.response_class = response_class

    def get_response_kwargs(self, exception):
        raise NotImplementedError

    def get_response(self, exception):
        return self.response_class(**self.get_response_kwargs(exception))


class ResponseErrorFactory(ResponseFactory):

    def __init__(self, msg, code, response_class=RESTErrorResponse):
        super(ResponseErrorFactory, self).__init__(response_class)
        self.msg = msg
        self.code = code

    def get_response_kwargs(self, exception):
        return {
            'msg': self.msg,
            'code': self.code,
        }


class ResponseExceptionFactory(ResponseFactory):

    def get_response_kwargs(self, exception):
        return {
            'msg': exception.message,
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-07-29 21:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('app', '0002_user_manual_created_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='estimate_minutes',
            field=models.IntegerField(blank=True, null=True, verbose_name='logged minutes'),
        ),
        migrations.AddField(
            model_name='issue',
            name='logged_minutes',
            field=models.IntegerField(blank=True, null=True, verbose_name='logged minutes'),
        ),
    ]



from django.utils.encoding import force_text
from django.db import models
from django.db.models.fields import FieldDoesNotExist
from django.core.exceptions import ObjectDoesNotExist
from django.template.defaultfilters import capfirst
from django.forms.forms import pretty_name

from chamber.utils import get_class_method

from collections import OrderedDict

from pyston.utils import split_fields, is_match, get_model_from_relation_or_none, LOOKUP_SEP
from pyston.utils.compatibility import get_all_related_objects_from_model, get_concrete_field, get_model_from_relation


class Field:

    def __init__(self, key_path, label_path):
        self.key_path = key_path
        self.label_path = label_path

    def __str__(self):
        return capfirst(' '.join(map(force_text, self.label_path))).strip()

    def __hash__(self):
        return hash(LOOKUP_SEP.join(self.key_path))

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __ne__(self, other):
        return not self.__eq__(other)


class FieldsetGenerator:

    def __init__(self, resource=None, fields_string=None):
        self.resource = resource
        self.fields_string = fields_string

    def _get_resource_class(self, obj):
        from pyston.resource import typemapper

        resource_class = typemapper.get(obj)
        if resource_class:
            return resource_class

    def _get_field_label_from_model_related_objects(self, model, field_name):
        for rel in get_all_related_objects_from_model(model):
            reverse_name = rel.get_accessor_name()
            if field_name == reverse_name:
                model = get_model_from_relation(model, field_name)
                if isinstance(rel.field, models.OneToOneField):
                    return model._meta.verbose_name
                else:
                    return model._meta.verbose_name_plural
        return None

    def _get_field_label_from_resource_or_model_method(self, resource_or_model, field_name):
        return get_class_method(resource_or_model, field_name).short_description

    def _get_field_label_from_model_field(self, model, field_name):
        return get_concrete_field(model, field_name).verbose_name

    def _get_field_label_from_model(self, model, resource, field_name):
        try:
            return self._get_field_label_from_model_field(model, field_name)
        except FieldDoesNotExist:
            if resource:
                resource_and_model = (resource, model)
            else:
                resource_and_model = (model,)

            for resource_or_model in resource_and_model:
                try:
                    return self._get_field_label_from_resource_or_model_method(resource_or_model, field_name)
                except (AttributeError, ObjectDoesNotExist):
                    pass

            return self._get_field_label_from_model_related_objects(model, field_name) or pretty_name(field_name)

    def _get_label(self, field_name, model):
        if model:
            return (
                self._get_field_label_from_model(model, self._get_resource_class(model), field_name)
                if field_name != '_obj_name' and field_name else ''
            )
        else:
            return field_name

    def _recursive_generator(self, fields, fields_string, model=None, key_path=None, label_path=None):
        key_path = key_path or []
        label_path = label_path or []

        if not fields_string:
            fields.append(Field(key_path, label_path))
        else:
            for field in split_fields(fields_string):
                if is_match('^[^\(\)]+\(.+\)$', field):
                    field_name, subfields_string = field[:len(field) - 1].split('(', 1)
                else:
                    field_name = field
                    subfields_string = None

                if LOOKUP_SEP in field_name:
                    field_name, subfields_string = field.split(LOOKUP_SEP, 1)

                self._recursive_generator(fields, subfields_string, get_model_from_relation_or_none(model, field_name),
                                          key_path + [field_name],
                                          label_path + [self._get_label(field_name, model)])

    def generate(self):
        fields = []
        self._recursive_generator(fields, self.fields_string, getattr(self.resource, 'model', None))
        return fields


class DataFieldset:

    def __init__(self, data):
        self.root = {}
        # OrderedDict is used as OrderedSet
        self.fieldset = OrderedDict()
        self._init_data(data)

    def _tree_contains(self, key_path):
        current = self.root.get(key_path[0])

        if current is None:
            return False

        for key in key_path[1:]:
            if not current:
                return True
            elif key not in current.keys():
                return False
            else:
                current = current.get(key)

        return not bool(current)

    def _remove_childs(self, key_path, tree):
        if not tree:
            del self.fieldset[LOOKUP_SEP.join(key_path)]
        else:
            for key, subtree in tree.items():
                self._remove_childs(key_path + [key], subtree)

    def _add(self, key_path):
        if not self._tree_contains(key_path):
            current = self.root
            for key in key_path:
                current[key] = current.get(key, {})
                prev = current
                current = current[key]

            if current:
                self._remove_childs(key_path, current)

            prev[key] = {}
            self.fieldset[LOOKUP_SEP.join(key_path)] = None

    def _init_data(self, converted_data, key_path=None):
        key_path = key_path or []

        if isinstance(converted_data, dict):
            for key, val in converted_data.items():
                self._init_data(val, list(key_path) + [key])
        elif isinstance(converted_data, (list, tuple, set)):
            is_last_list = False
            for val in converted_data:
                if isinstance(list, (list, tuple, set)):
                    is_last_list = True
                    break
                self._init_data(val, list(key_path))
            if is_last_list:
                self._add(key_path)
        elif converted_data is not None:
            self._add(key_path)

    def __iter__(self):
        return iter(self.fieldset.keys())

    def __nonzero__(self):
        return bool(self.fieldset)

    def __str__(self):
        return '%s' % ','.join(self.fieldset.keys())

    def __len__(self):
        return len(self.fieldset)



import os
import decimal
import datetime
import inspect
import mimetypes
import types

from collections import OrderedDict

from django.db.models import Model
from django.db.models.query import QuerySet
from django.db.models.fields.files import FileField

try:
    from django.db.models.fields.related import ForeignRelatedObjectsDescriptor, SingleRelatedObjectDescriptor
except ImportError:
    from django.db.models.fields.related import (ReverseManyToOneDescriptor as ForeignRelatedObjectsDescriptor,
                                                 ReverseOneToOneDescriptor as SingleRelatedObjectDescriptor)

from django.utils import formats, timezone
from django.utils.encoding import force_text
from django.utils.translation import ugettext
from django.utils.html import conditional_escape

from chamber.utils.datastructures import Enum
from chamber.utils import get_class_method

from .conf import settings
from .exception import UnsupportedMediaTypeException, NotAllowedException
from .utils import rfs
from .utils.compatibility import get_reverse_field_name, get_last_parent_pk_field_name
from .utils.helpers import QuerysetIteratorHelper, UniversalBytesIO, serialized_data_to_python, str_to_class
from .converters import get_converter
from .forms import RESTDictError, RESTListError, RESTDictIndexError


default_serializers = []


class Serializable:

    def serialize(self, serialization_format, **kwargs):
        raise NotImplementedError


class SerializableObj(Serializable):

    resource_typemapper = {}

    def _get_value(self, field, serialization_format, request, **kwargs):
        val = getattr(self, field)
        return get_serializer(
            val, request=request, resource_typemapper=self.resource_typemapper
        ).serialize(val, serialization_format, **kwargs)

    def serialize(self, serialization_format, request=None, **kwargs):
        return {field_name: self._get_value(field_name, serialization_format, request, **kwargs)
                for field_name in self.get_fields()}

    def get_fields(self):
        return self.RESTMeta.fields


class SerializationException(Exception):
    pass


def register(serialized_types):
    """
    Adds throttling validator to a class.
    """
    def _register(klass):
        if klass not in (serializer for _, serializer in default_serializers):
            default_serializers.insert(0, (serialized_types, klass))
        return klass
    return _register


def get_resource_class_or_none(thing, resource_typemapper=None):
    from .resource import typemapper as global_resource_typemapper

    resource_typemapper = {} if resource_typemapper is None else resource_typemapper
    resource_class = resource_typemapper.get(thing) or global_resource_typemapper.get(thing)
    if isinstance(resource_class, str):
        resource_class = str_to_class(resource_class)
    return resource_class


def get_resource_or_none(request, thing, resource_typemapper=None):
    from .resource import typemapper as global_resource_typemapper

    resource_class = get_resource_class_or_none(thing, resource_typemapper)
    return resource_class(request) if resource_class else None


def get_serializer(thing, request=None, resource_typemapper=None):
    if request:
        thing_class = thing.model if isinstance(thing, QuerySet) else type(thing)
        resource = get_resource_or_none(request, thing_class, resource_typemapper)
        if resource:
            return resource.serializer(resource, request=request)

    for serialized_types, serializer in default_serializers:
        if isinstance(thing, serialized_types):
            return serializer(request=request)
    return DefaultSerializer(request=request)


class RawVerboseValue:
    """
    Return RAW, VERBOSE or BOTH values according to serialization type
    """

    def __init__(self, raw_value, verbose_value):
        self.raw_value = raw_value
        self.verbose_value = verbose_value

    def get_value(self, serialization_format):
        if serialization_format == Serializer.SERIALIZATION_TYPES.RAW:
            return self.raw_value
        elif serialization_format == Serializer.SERIALIZATION_TYPES.VERBOSE:
            return self.verbose_value
        elif self.raw_value == self.verbose_value:
            return self.raw_value
        else:
            return {'_raw': self.raw_value, '_verbose': self.verbose_value}


class LazySerializedData:

    def __init__(self, serializer, data, serialization_format, **kwargs):
        self.serializer = serializer
        self.data = data
        self.serialization_format = serialization_format
        self.kwargs = kwargs

    def serialize(self):
        return self.serializer.serialize(self.data, self.serialization_format, **self.kwargs)


class LazyMappedSerializedData:

    def __init__(self, data, data_mapping):
        self.data = data
        self.data_mapping = data_mapping

    def _map_key(self, lookup_key):
        return self.data_mapping.get(lookup_key, lookup_key)

    def serialize(self):
        if isinstance(self.data, RESTDictError):
            return RESTDictError({self._map_key(key): val for key, val in self.data.items()})
        elif isinstance(self.data, RESTListError):
            return RESTListError([LazyMappedSerializedData(val, self.data_mapping) for val in self.data])
        elif isinstance(self.data, RESTDictIndexError):
            return RESTDictIndexError(
                self.data.index,
                {self._map_key(key): val for key, val in self.data.data.items()}
            )
        elif isinstance(self.data, (types.GeneratorType, list, tuple)):
            return [LazyMappedSerializedData(val, self.data_mapping) for val in self.data]
        elif isinstance(self.data, dict):
            return OrderedDict(((self._map_key(key), val) for key, val in self.data.items()))
        else:
            return self.data


LAZY_SERIALIZERS = (LazySerializedData, LazyMappedSerializedData)


class Serializer:
    """
    REST serializer and deserializer, firstly is data serialized to standard python data types and after that is
    used convertor for final serialization
    """

    SERIALIZATION_TYPES = Enum('VERBOSE', 'RAW', 'BOTH')

    def __init__(self, request=None):
        self.request = request

    def _get_serializer(self, data):
        return get_serializer(data, request=self.request)

    def _data_to_python(self, data, serialization_format, lazy=False, **kwargs):
        return self._get_serializer(data).serialize(data, serialization_format, **kwargs)

    def _lazy_data_to_python(self, data, serialization_format, lazy=False, **kwargs):
        if lazy:
            return LazySerializedData(
                self._get_serializer(data), data, serialization_format, lazy=lazy, **kwargs
            )
        else:
            return self._data_to_python(data, serialization_format, lazy=lazy, **kwargs)

    def serialize(self, data, serialization_format, **kwargs):
        raise NotImplementedError

    def deserialize(self, data):
        raise NotImplementedError


class ResourceSerializerMixin:

    def __init__(self, resource, request=None):
        self.resource = resource
        super(ResourceSerializerMixin, self).__init__(request=request)

    def _get_serializer(self, data):
        return get_serializer(data, request=self.request, resource_typemapper=self.resource.resource_typemapper)

    def deserialize(self, data):
        return data

    def _serialize_recursive(self, data, serialization_format, **kwargs):
        if isinstance(data, dict):
            return dict(
                [
                    (k, self._serialize_recursive(v, serialization_format, **kwargs)) for k, v in data.items()
                ]
            )
        elif isinstance(data, (list, tuple, set)):
            return (self._serialize_recursive(v, serialization_format, **kwargs) for v in data)
        else:
            return self._serialize_other(data, serialization_format, **kwargs)

    def _serialize_other(self, data, serialization_format, **kwargs):
        copy_kwargs = kwargs.copy()
        copy_kwargs['via'] = self.resource._get_via(copy_kwargs.get('via'))
        return self._data_to_python(data, serialization_format, **copy_kwargs)

    def serialize(self, data, serialization_format, **kwargs):
        return self._serialize_recursive(data, serialization_format, **kwargs)


class ResourceSerializer(ResourceSerializerMixin, Serializer):

    def _serialize_recursive(self, data, serialization_format, **kwargs):
        if isinstance(data, dict):
            return self.resource.update_serialized_data(
                super(ResourceSerializer, self)._serialize_recursive(data, serialization_format, **kwargs)
            )
        else:
            return super(ResourceSerializer, self)._serialize_recursive(data, serialization_format, **kwargs)


class ObjectResourceSerializer(ResourceSerializerMixin, Serializer):

    def _serialize_recursive(self, data, serialization_format, **kwargs):
        if isinstance(data, self.resource.model):
            if not self.resource.has_read_obj_permission(obj=data, via=kwargs.get('via')):
                raise NotAllowedException
            return self.resource.update_serialized_data(
                data.serialize(serialization_format, request=self.request, **kwargs)
            )
        else:
            return super(ObjectResourceSerializer, self)._serialize_recursive(data, serialization_format, **kwargs)


@register(str)
class StringSerializer(Serializer):

    def serialize(self, data, serialization_format, allow_tags=False, **kwargs):
        serialized_string = force_text(data, strings_only=True)
        return serialized_string if allow_tags or settings.ALLOW_TAGS else conditional_escape(serialized_string)


class DefaultSerializer(Serializer):

    def serialize(self, data, serialization_format, **kwargs):
        return force_text(data, strings_only=True)


@register(datetime.datetime)
class DateTimeSerializer(Serializer):

    def serialize(self, data, serialization_format, **kwargs):
        return timezone.localtime(data)


@register(dict)
class DictSerializer(Serializer):

    def serialize(self, data, serialization_format, requested_fieldset=None,
                  extended_fieldset=None, exclude_fields=None, **kwargs):
        return dict([(k, self._data_to_python(v, serialization_format, **kwargs))
                     for k, v in data.items()])


@register((list, tuple, set))
class CollectionsSerializer(Serializer):

    def serialize(self, data, serialization_format, requested_fieldset=None,
                   extended_fieldset=None, exclude_fields=None, **kwargs):
        return (self._data_to_python(v, serialization_format, **kwargs) for v in data)


@register(decimal.Decimal)
class DecimalSerializer(Serializer):

    def serialize(self, data, serialization_format, **kwargs):
        return data


@register(RawVerboseValue)
class RawVerboseSerializer(Serializer):

    def serialize(self, data, serialization_format, lazy=False, **kwargs):
        return self._data_to_python(data.get_value(serialization_format), serialization_format, lazy=False, **kwargs)


@register(Serializable)
class SerializableSerializer(Serializer):

    def serialize(self, data, serialization_format, **kwargs):
        return data.serialize(serialization_format, request=self.request, **kwargs)


@register((Model, QuerySet, QuerysetIteratorHelper))
class ModelSerializer(Serializer):

    RESERVED_FIELDS = {'read', 'update', 'create', 'delete', 'model', 'allowed_methods', 'fields', 'exclude'}

    def _get_resource_method_fields(self, resource, fields):
        out = {}
        for field in fields.flat() - self.RESERVED_FIELDS:
            t = getattr(resource, str(field), None)
            if t and callable(t):
                out[field] = t
        return out

    def _get_model_fields(self, obj):
        out = {}
        for f in obj._meta.fields:
            if hasattr(f, 'serialize') and f.serialize:
                out[f.name] = f
        return out

    def _get_m2m_fields(self, obj):
        out = {}
        for mf in obj._meta.many_to_many:
            if mf.serialize:
                out[mf.name] = mf
        return out

    def _get_verbose_value(self, raw, field_or_method, obj, **kwargs):
        if hasattr(field_or_method, 'humanized') and field_or_method.humanized:
            return field_or_method.humanized(raw, obj, **kwargs)
        elif hasattr(field_or_method, 'choices') and field_or_method.choices:
            return getattr(obj, 'get_{}_display'.format(field_or_method.attname))()
        if isinstance(raw, bool):
            return raw and ugettext('Yes') or ugettext('No')
        elif isinstance(raw, datetime.datetime):
            return formats.localize(timezone.template_localtime(raw))
        elif isinstance(raw, (datetime.date, datetime.time)):
            return formats.localize(raw)
        else:
            return raw

    def _value_to_raw_verbose(self, val, field_or_method, obj, **kwargs):
        return RawVerboseValue(val, self._get_verbose_value(val, field_or_method, obj, **kwargs))

    def _method_to_python(self, method, obj, serialization_format, allow_tags=False, **kwargs):
        method_kwargs_names = inspect.getargspec(method)[0][1:]

        method_kwargs = {}

        fun_kwargs = {'request': self.request, 'obj': obj} if self.request else {'obj': obj}

        for arg_name in method_kwargs_names:
            if arg_name in fun_kwargs:
                method_kwargs[arg_name] = fun_kwargs[arg_name]

        if len(method_kwargs_names) == len(method_kwargs):
            return self._data_to_python(
                self._value_to_raw_verbose(method(**method_kwargs), method, obj,
                                           **{k: v for k, v in method_kwargs.items() if k != 'obj'}),
                serialization_format, allow_tags=allow_tags or getattr(method, 'allow_tags', False), **kwargs
            )
        else:
            raise SerializationException('Invalid method parameters')

    def _model_field_to_python(self, field, obj, serialization_format, allow_tags=False, **kwargs):
        return (self._lazy_data_to_python if field.is_relation else self._data_to_python)(
            self._value_to_raw_verbose(self._get_model_field_raw_value(obj, field), field, obj)
            if not field.remote_field else getattr(obj, field.name),
            serialization_format, allow_tags=allow_tags or getattr(field, 'allow_tags', False), **kwargs
        )

    def _m2m_field_to_python(self, field, obj, serialization_format, allow_tags=False, **kwargs):
        return (self._data_to_python(m, serialization_format,
                                     allow_tags=allow_tags or getattr(field, 'allow_tags', False), **kwargs)
                for m in getattr(obj, field.name).all())

    def _get_reverse_excluded_fields(self, field, obj):
        model = obj.__class__
        exclude_fields = []
        if hasattr(model, field) and isinstance(getattr(model, field, None),
                                                (ForeignRelatedObjectsDescriptor, SingleRelatedObjectDescriptor)):
            exclude_fields.append(get_reverse_field_name(model, field))

        return exclude_fields

    def _reverse_qs_to_python(self, val, field, obj, serialization_format, **kwargs):
        kwargs['exclude_fields'] = self._get_reverse_excluded_fields(field, obj)
        return (self._data_to_python(m, serialization_format, **kwargs) for m in val.all())

    def _reverse_to_python(self, val, field, obj, serialization_format, **kwargs):
        kwargs['exclude_fields'] = self._get_reverse_excluded_fields(field, obj)
        return self._lazy_data_to_python(val, serialization_format, **kwargs)

    def _copy_kwargs(self, resource, kwargs):
        subkwargs = kwargs.copy()
        subkwargs['exclude_fields'] = None
        subkwargs['via'] = resource._get_via(kwargs.get('via')) if resource else kwargs.get('via')
        return subkwargs

    def _get_field_name(self, field, requested_field, subkwargs):
        if field.subfieldset:
            field_name, subkwargs['extended_fieldset'] = field.name, field.subfieldset
        else:
            field_name, subkwargs['extended_fieldset'] = field.name, None

        if requested_field and requested_field.subfieldset:
            subkwargs['requested_fieldset'] = requested_field.subfieldset
        elif field.subfieldset:
            subkwargs['requested_fieldset'] = field.subfieldset

        return field_name

    def _get_file_field_value(self, val):
        if val:
            filename = os.path.basename(val.name)
            return {
                'filename': filename,
                'content_type': (
                    mimetypes.types_map.get('.{}'.format(filename.split('.')[-1])) if '.' in filename else None
                ),
                'url': val.url,
            }
        else:
            return None

    def _get_model_field_raw_value(self, obj, field):
        val = getattr(obj, field.attname)
        return self._get_file_field_value(val) if isinstance(field, FileField) else val

    def _field_to_python(self, field_name, resource_method_fields, model_fields, m2m_fields,
                         obj, serialization_format, allow_tags=False, **kwargs):

        if field_name == '_obj_name':
            return force_text(obj)
        elif field_name in resource_method_fields:
            return self._method_to_python(resource_method_fields[field_name], obj, serialization_format,
                                          allow_tags=allow_tags, **kwargs)
        elif field_name in m2m_fields:
            return self._m2m_field_to_python(
                m2m_fields[field_name], obj, serialization_format, allow_tags=allow_tags, **kwargs
            )
        elif field_name in model_fields:
            return self._model_field_to_python(
                model_fields[field_name], obj, serialization_format, allow_tags=allow_tags, **kwargs
            )
        else:
            val = getattr(obj, field_name, None) if hasattr(obj, field_name) else None
            if hasattr(val, 'all'):
                return self._reverse_qs_to_python(
                    val, field_name, obj, serialization_format, allow_tags=allow_tags, **kwargs
                )
            elif isinstance(val, Model):
                return self._reverse_to_python(
                    val, field_name, obj, serialization_format, allow_tags=allow_tags, **kwargs
                )
            elif callable(val):
                return self._method_to_python(val, obj, serialization_format, allow_tags=allow_tags, **kwargs)
            else:
                method = get_class_method(obj, field_name)
                return self._data_to_python(
                    self._value_to_raw_verbose(val, method, obj),
                    serialization_format,
                    allow_tags=allow_tags or method is not None and getattr(method, 'allow_tags', False),
                    **kwargs
                )

    def _fields_to_python(self, obj, serialization_format, fieldset, requested_fieldset, **kwargs):
        model_resource = self._get_model_resource(obj)
        resource_method_fields = self._get_resource_method_fields(model_resource, fieldset)
        model_fields = self._get_model_fields(obj)
        m2m_fields = self._get_m2m_fields(obj)

        out = OrderedDict()

        for field in fieldset.fields:
            subkwargs = self._copy_kwargs(model_resource, kwargs)
            requested_field = None
            if requested_fieldset:
                requested_field = requested_fieldset.get(field.name)
            field_name = self._get_field_name(field, requested_field, subkwargs)
            out[field_name] = self._field_to_python(
                field_name, resource_method_fields, model_fields, m2m_fields, obj, serialization_format, **subkwargs
            )
        return out

    def _get_model_resource(self, obj):
        return None

    def _get_fieldset_from_resource(self, model_resource, obj, via, has_read_permission):
        if not has_read_permission:
            return model_resource.get_guest_fields_rfs(obj)
        else:
            return model_resource.get_general_fields_rfs(obj)

    def _get_allowed_fieldset_from_resource(self, model_resource, obj, via, has_read_permission):
        if not has_read_permission:
            return model_resource.get_guest_fields_rfs(obj)
        else:
            return model_resource.get_allowed_fields_rfs(obj)

    def _get_direct_serialization_fields(self, obj):
        return rfs(obj._rest_meta.direct_serialization_fields).join(rfs(obj._rest_meta.default_fields))

    def _get_fieldset(self, obj, extended_fieldset, requested_fieldset, exclude_fields, via, direct_serialization,
                      serialized_objects):

        if self._get_obj_serialization_name(obj) in serialized_objects:
            return rfs((get_last_parent_pk_field_name(obj),))

        model_resource = self._get_model_resource(obj)

        if model_resource:
            has_read_permission = model_resource.has_read_obj_permission(obj=obj, via=via)
            default_fieldset = self._get_fieldset_from_resource(model_resource, obj, via, has_read_permission)
            allowed_fieldset = self._get_allowed_fieldset_from_resource(model_resource, obj, via, has_read_permission)
        else:
            direct_serialization_fields = self._get_direct_serialization_fields(obj)
            allowed_fieldset = rfs(
                requested_fieldset if requested_fieldset else (
                    direct_serialization_fields if direct_serialization else obj._rest_meta.guest_fields
                )
            )
            default_fieldset = rfs(direct_serialization_fields if direct_serialization else obj._rest_meta.guest_fields)

        if extended_fieldset:
            default_fieldset.join(extended_fieldset)
            allowed_fieldset.join(extended_fieldset)

        if requested_fieldset:
            # requested_fieldset must be cloned because RFS is not immutable and intersection change it
            fieldset = rfs(requested_fieldset).intersection(allowed_fieldset)
        else:
            fieldset = default_fieldset.intersection(allowed_fieldset)

        if exclude_fields:
            fieldset.subtract(exclude_fields)
        return fieldset

    def _get_obj_serialization_name(self, obj):
        return '{}__{}'.format(obj._meta.db_table, obj.pk)

    def _obj_to_python(self, obj, serialization_format, requested_fieldset=None, extended_fieldset=None,
                       exclude_fields=None, direct_serialization=False,
                       serialized_objects=None, **kwargs):
        exclude_fields = [] if exclude_fields is None else exclude_fields
        serialized_objects = set() if serialized_objects is None else set(serialized_objects)
        fieldset = self._get_fieldset(obj, extended_fieldset, requested_fieldset, exclude_fields,
                                      kwargs.get('via'), direct_serialization, serialized_objects)
        serialized_objects.add(self._get_obj_serialization_name(obj))
        return self._fields_to_python(obj, serialization_format, fieldset, requested_fieldset,
                                      serialized_objects=serialized_objects,
                                      direct_serialization=direct_serialization, **kwargs)

    def serialize(self, data, serialization_format, **kwargs):
        if isinstance(data, QuerysetIteratorHelper):
            return (self._obj_to_python(obj, serialization_format, **kwargs) for obj in data.iterator())
        elif isinstance(data, QuerySet):
            return (self._obj_to_python(obj, serialization_format, **kwargs) for obj in data)
        elif isinstance(data, Model):
            return self._obj_to_python(data, serialization_format, **kwargs)
        else:
            raise NotImplementedError


class ModelResourceSerializer(ResourceSerializerMixin, ModelSerializer):

    def _get_model_resource(self, obj):
        return self.resource

    def _serialize_recursive(self, data, serialization_format, **kwargs):
        if (isinstance(data, self.resource.model) or
              isinstance(data, (QuerysetIteratorHelper, QuerySet)) and issubclass(data.model, self.resource.model)):
            return self.resource.update_serialized_data(
                ModelSerializer.serialize(self, data, serialization_format, **kwargs)
            )
        else:
            return super(ModelResourceSerializer, self)._serialize_recursive(data, serialization_format, **kwargs)


def serialize(data, requested_fieldset=None, serialization_format=Serializer.SERIALIZATION_TYPES.RAW,
              converter_name=None, allow_tags=None):
    from pyston.converters import get_default_converter_name

    converter_name = converter_name if converter_name is not None else get_default_converter_name()
    try:
        converter = get_converter(converter_name) if converter_name != 'python' else None
    except ValueError:
        raise UnsupportedMediaTypeException

    requested_fieldset = rfs(requested_fieldset) if requested_fieldset is not None else None
    converted_dict = get_serializer(data).serialize(
        data, serialization_format, requested_fieldset=requested_fieldset, direct_serialization=True,
        allow_tags=allow_tags or (converter and converter.allow_tags)
    )
    if converter:
        os = UniversalBytesIO()
        converter.encode_to_stream(os, converted_dict)
        return os.get_string_value()
    else:
        return serialized_data_to_python(converted_dict)



from germanium.tools.trivials import assert_equal
from germanium.tools.http import assert_http_method_not_allowed
from germanium.tools.rest import assert_valid_JSON_response

from .test_case import PystonTestCase
from .factories import UserFactory, IssueFactory


class ExtraResourceTestCase(PystonTestCase):

    def test_not_supported_message_for_put_post_and_delete(self):
        resp = self.put(self.EXTRA_API_URL, data={})
        assert_http_method_not_allowed(resp)

        resp = self.post(self.EXTRA_API_URL, data={})
        assert_http_method_not_allowed(resp)

        resp = self.delete(self.EXTRA_API_URL)
        assert_http_method_not_allowed(resp)

    def test_should_return_data_for_get(self):
        resp = self.get(self.EXTRA_API_URL)
        assert_valid_JSON_response(resp)

    def test_resource_with_serializable_result(self):
        [IssueFactory(solver=UserFactory()) for _ in range(10)]

        resp_data = self.deserialize(self.get(self.COUNT_ISSUES_PER_USER))
        assert_equal(len(resp_data), 10 * 3)
        for row in resp_data:
            assert_equal(set(row.keys()), {'email', 'created_issues_count'})

    def test_resource_with_serializableobj_result(self):
        issues = [IssueFactory() for _ in range(10)]
        for issue in issues:
            issue.watched_by.add(*[UserFactory() for _ in range(5)])

        resp_data = self.deserialize(self.get(self.COUNT_WATCHERS_PER_ISSUE))
        assert_equal(len(resp_data), 10)
        for row in resp_data:
            assert_equal(set(row.keys()), {'name', 'watchers_count'})
            assert_equal(row['watchers_count'], 5)



from setuptools import setup, find_packages

from pyston.version import get_version


setup(
    name="django-pyston",
    version=get_version(),
    description="Pyston is a Django mini-framework creating APIs.",
    author='Lubos Matl',
    author_email='matllubos@gmail.com',
    url='https://github.com/matllubos/django-pyston',
    license='BSD',
    package_dir={'pyston': 'pyston'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'django>=1.10',
        'python-mimeparse==0.1.4',
        'django-chamber>=0.4.0',
        'pyparsing>=2.2.0',
    ],
    zip_safe=False
)



from unittest.case import TestCase

from django.core.exceptions import FieldError

from germanium.tools import assert_true, assert_false, assert_raises, assert_equal, assert_is_none

from pyston.utils.compatibility import (
    is_relation, is_one_to_one, is_many_to_many, is_many_to_one, is_reverse_many_to_many, is_reverse_many_to_one,
    is_reverse_one_to_one, get_model_from_relation, get_model_from_relation_or_none, get_reverse_field_name
)

from app.models import Issue, User


class CompatibilityTestCase(TestCase):

    def test_is_relation(self):
        assert_true(is_relation(Issue, 'watched_by'))
        assert_true(is_relation(Issue, 'created_by'))
        assert_true(is_relation(Issue, 'solver'))
        assert_true(is_relation(Issue, 'leader'))
        assert_false(is_relation(Issue, 'name'))
        assert_false(is_relation(Issue, 'created_at'))
        assert_false(is_relation(Issue, 'invalid'))

        assert_true(is_relation(User, 'watched_issues'))
        assert_true(is_relation(User, 'created_issues'))
        assert_true(is_relation(User, 'solving_issue'))
        assert_true(is_relation(User, 'leading_issue'))

    def test_is_one_to_one(self):
        assert_false(is_one_to_one(Issue, 'watched_by'))
        assert_false(is_one_to_one(Issue, 'created_by'))
        assert_true(is_one_to_one(Issue, 'solver'))
        assert_true(is_one_to_one(Issue, 'leader'))
        assert_false(is_one_to_one(Issue, 'name'))
        assert_false(is_one_to_one(Issue, 'created_at'))
        assert_false(is_one_to_one(Issue, 'invalid'))

        assert_false(is_one_to_one(User, 'watched_issues'))
        assert_false(is_one_to_one(User, 'created_issues'))
        assert_false(is_one_to_one(User, 'solving_issue'))
        assert_false(is_one_to_one(User, 'leading_issue'))

    def test_is_many_to_one(self):
        assert_false(is_many_to_one(Issue, 'watched_by'))
        assert_true(is_many_to_one(Issue, 'created_by'))
        assert_false(is_many_to_one(Issue, 'solver'))
        assert_false(is_many_to_one(Issue, 'leader'))
        assert_false(is_many_to_one(Issue, 'name'))
        assert_false(is_many_to_one(Issue, 'created_at'))
        assert_false(is_many_to_one(Issue, 'invalid'))

        assert_false(is_many_to_one(User, 'watched_issues'))
        assert_false(is_many_to_one(User, 'created_issues'))
        assert_false(is_many_to_one(User, 'solving_issue'))
        assert_false(is_many_to_one(User, 'leading_issue'))

    def test_is_many_to_many(self):
        assert_true(is_many_to_many(Issue, 'watched_by'))
        assert_false(is_many_to_many(Issue, 'created_by'))
        assert_false(is_many_to_many(Issue, 'solver'))
        assert_false(is_many_to_many(Issue, 'leader'))
        assert_false(is_many_to_many(Issue, 'name'))
        assert_false(is_many_to_many(Issue, 'created_at'))
        assert_false(is_many_to_many(Issue, 'invalid'))

        assert_false(is_many_to_many(User, 'watched_issues'))
        assert_false(is_many_to_many(User, 'created_issues'))
        assert_false(is_many_to_many(User, 'solving_issue'))
        assert_false(is_many_to_many(User, 'leading_issue'))

    def test_is_reverse_one_to_one(self):
        assert_false(is_reverse_one_to_one(Issue, 'watched_by'))
        assert_false(is_reverse_one_to_one(Issue, 'created_by'))
        assert_false(is_reverse_one_to_one(Issue, 'solver'))
        assert_false(is_reverse_one_to_one(Issue, 'leader'))
        assert_false(is_reverse_one_to_one(Issue, 'name'))
        assert_false(is_reverse_one_to_one(Issue, 'created_at'))
        assert_false(is_reverse_one_to_one(Issue, 'invalid'))

        assert_false(is_reverse_one_to_one(User, 'watched_issues'))
        assert_false(is_reverse_one_to_one(User, 'created_issues'))
        assert_true(is_reverse_one_to_one(User, 'solving_issue'))
        assert_true(is_reverse_one_to_one(User, 'leading_issue'))

    def test_is_reverse_many_to_one(self):
        assert_false(is_reverse_many_to_one(Issue, 'watched_by'))
        assert_false(is_reverse_many_to_one(Issue, 'created_by'))
        assert_false(is_reverse_many_to_one(Issue, 'solver'))
        assert_false(is_reverse_many_to_one(Issue, 'leader'))
        assert_false(is_reverse_many_to_one(Issue, 'name'))
        assert_false(is_reverse_many_to_one(Issue, 'created_at'))
        assert_false(is_reverse_many_to_one(Issue, 'invalid'))

        assert_false(is_reverse_many_to_one(User, 'watched_issues'))
        assert_true(is_reverse_many_to_one(User, 'created_issues'))
        assert_false(is_reverse_many_to_one(User, 'solving_issue'))
        assert_false(is_reverse_many_to_one(User, 'leading_issue'))

    def test_is_reverse_many_to_many(self):
        assert_false(is_reverse_many_to_many(Issue, 'watched_by'))
        assert_false(is_reverse_many_to_many(Issue, 'created_by'))
        assert_false(is_reverse_many_to_many(Issue, 'solver'))
        assert_false(is_reverse_many_to_many(Issue, 'leader'))
        assert_false(is_reverse_many_to_many(Issue, 'name'))
        assert_false(is_reverse_many_to_many(Issue, 'created_at'))
        assert_false(is_reverse_many_to_many(Issue, 'invalid'))

        assert_true(is_reverse_many_to_many(User, 'watched_issues'))
        assert_false(is_reverse_many_to_many(User, 'created_issues'))
        assert_false(is_reverse_many_to_many(User, 'solving_issue'))
        assert_false(is_reverse_many_to_many(User, 'leading_issue'))

    def test_get_model_from_relation(self):
        assert_equal(get_model_from_relation(Issue, 'watched_by'), User)
        assert_equal(get_model_from_relation(Issue, 'created_by'), User)
        assert_equal(get_model_from_relation(Issue, 'solver'), User)
        assert_equal(get_model_from_relation(Issue, 'leader'), User)
        assert_raises(FieldError, get_model_from_relation, Issue, 'name')
        assert_raises(FieldError, get_model_from_relation, Issue, 'created_at')
        assert_raises(FieldError, get_model_from_relation, Issue, 'invalid')

        assert_equal(get_model_from_relation(User, 'watched_issues'), Issue)
        assert_equal(get_model_from_relation(User, 'created_issues'), Issue)
        assert_equal(get_model_from_relation(User, 'solving_issue'), Issue)
        assert_equal(get_model_from_relation(User, 'leading_issue'), Issue)

    def test_get_model_from_relation_or_none(self):
        assert_equal(get_model_from_relation_or_none(Issue, 'watched_by'), User)
        assert_equal(get_model_from_relation_or_none(Issue, 'created_by'), User)
        assert_equal(get_model_from_relation_or_none(Issue, 'solver'), User)
        assert_equal(get_model_from_relation_or_none(Issue, 'leader'), User)
        assert_is_none(get_model_from_relation_or_none(Issue, 'name'))
        assert_is_none(get_model_from_relation_or_none(Issue, 'created_at'))
        assert_is_none(get_model_from_relation_or_none(Issue, 'invalid'))

        assert_equal(get_model_from_relation_or_none(User, 'watched_issues'), Issue)
        assert_equal(get_model_from_relation_or_none(User, 'created_issues'), Issue)
        assert_equal(get_model_from_relation_or_none(User, 'solving_issue'), Issue)
        assert_equal(get_model_from_relation_or_none(User, 'leading_issue'), Issue)

    def test_get_reverse_field_name(self):
        assert_equal(get_reverse_field_name(Issue, 'watched_by'), 'watched_issues')
        assert_equal(get_reverse_field_name(Issue, 'created_by'), 'created_issues')
        assert_equal(get_reverse_field_name(Issue, 'solver'), 'solving_issue')
        assert_equal(get_reverse_field_name(Issue, 'leader'), 'leading_issue')
        assert_raises(FieldError, get_reverse_field_name, Issue, 'name')
        assert_raises(FieldError, get_reverse_field_name, Issue, 'created_at')
        assert_raises(FieldError, get_reverse_field_name, Issue, 'invalid')

        assert_equal(get_reverse_field_name(User, 'watched_issues'), 'watched_by')
        assert_equal(get_reverse_field_name(User, 'created_issues'), 'created_by')
        assert_equal(get_reverse_field_name(User, 'solving_issue'), 'solver')
        assert_equal(get_reverse_field_name(User, 'leading_issue'), 'leader')
